# Datové struktury (zlomky)

# Konstruktor:
def make_fract(num, den):
    return ["fract", num, den]

# Selektory:
def get_num(fract):
    return fract[1]

def get_den(fract):
    return fract[2]

"""
>>> fract = make_fract(1, 2)
>>> fract
['fract', 1, 2]
>>> get_num(fract)
1
>>> get_den(fract)
2
"""
def are_fracts_equal(fract1, fract2):
    num1 = get_num(fract1)
    den1 = get_den(fract1)
    num2 = get_num(fract2)
    den2 = get_den(fract2)
    return num1 * den2 == num2 * den1

"""
>>> are_fracts_equal(make_fract(1, 2), make_fract(2, 4))
True
"""
def add_fracts(fract1, fract2):
    num1 = get_num(fract1)
    den1 = get_den(fract1)
    num2 = get_num(fract2)
    den2 = get_den(fract2)
    return  make_fract(den2 * num1 + den1 * num2, den1 * den2)
        
"""
>>> add_fracts(make_fract(1, 2), make_fract(1, 3))
['fract', 5, 6]
>>> add_fracts(make_fract(1, 2), make_fract(1, 2))
['fract', 4, 4]
"""

def mult_fracts(fract1, fract2):
    num1 = get_num(fract1)
    den1 = get_den(fract1)
    num2 = get_num(fract2)
    den2 = get_den(fract2)
    return make_fract(num1 * num2, den1 * den2)

"""
>>> mult_fracts(make_fract(1, 2), make_fract(1, 3))
['fract', 1, 6]
"""
# greatest common divisor (gcd)
#
# Eukleidův algoritmus:
def get_gcd(x, y):
    while y != 0:
        tmp = x % y
        x = y
        y = tmp
    return x

"""
>>> get_gcd(6, 8)
2
"""

def reduce_fract(fract):
    num = get_num(fract)
    den = get_den(fract)
    gcd = get_gcd(num, den)
    return make_fract(num // gcd, den // gcd)

"""
>>> reduce_fract(make_fract(3, 6))
['fract', 1, 2]
>>> reduce_fract(make_fract(6, 8))
['fract', 3, 4]
"""
def invert_add(fract):
    return make_fract(-get_num(fract), get_den(fract))
"""
>>> invert_add(make_fract(1, 2))
['fract', -1, 2]
"""

def sub_fracts(fract1, fract2):
    return add_fracts(fract1, invert_add(fract2))

"""
>>> sub_fracts(make_fract(1, 2), make_fract(1, 3))
['fract', 1, 6]
"""

def invert_mult(fract):
    return make_fract(get_den(fract), get_num(fract))

"""
>>> invert_mult(make_fract(1, 2))
['fract', 2, 1]
"""

def div_fracts(fract1, fract2):
    return mult_fracts(fract1, invert_mult(fract2))

"""
>>> div_fracts(make_fract(1, 2), make_fract(2, 1))
['fract', 1, 4]
"""
def fract_to_integer(fract):
    return get_num(fract) // get_den(fract)
"""
>>> fract_to_integer(make_fract(10, 2))
5
>>> fract_to_integer(make_fract(10, 3))
3
"""
def add_zeros(string, count):
    for i in range(count):
        string = "0" + string
    return string

"""
>>> add_zeros("1", 2)
'001'
"""

def fract_to_decimal(fract, n):
    integer = fract_to_integer(fract)
    integer_part = str(integer)
    rest = sub_fracts(fract, make_fract(integer, 1))
    mult_rest = mult_fracts(rest, make_fract(10 ** n, 1))
    rest_int = fract_to_integer(mult_rest)
    str_rest = str(rest_int)
    # Musíme doplnit případné nuly:
    fract_part = add_zeros(str_rest, n - len(str_rest))
    # print(rest, mult_rest, fract_part)
    return integer_part + "." + fract_part

"""
>>> fract_to_decimal(make_fract(10, 2), 5)
'5.0'
>>> fract_to_decimal(make_fract(1, 3), 5)
"0.33333"
"""

def square(fract):
    return mult_fracts(fract, fract)

# square root (sqrt)
#
# Metoda tečen (Newtonova metoda)
def sqrt(fract, n):
    res = fract
    for i in range(n):
        res = mult_fracts(make_fract(1, 2),
                          add_fracts(res,
                                     div_fracts(fract, res)))
        print("Iterace:", i + 1)
        print("Odhad druhé odmocniny:", fract_to_decimal(res, 30))
        print("Druhá mocnina odhadu:", fract_to_decimal(square(res), 30))
    return res

# Jak získat přibližně druhou odmocninu 
# se zárukou správných několik (zadaných) cifer za desetinou čárkou?


# Úkoly na procvičení.
# Body roviny reprezentujeme poli ["point", x, y]
# a vektory poli ["vect", x, y].
#
# Jsou dány konstruktory a selektory:

def make_point(x, y):
    return ["point", x, y]

def get_point_x(point):
    return point[1]

def get_point_y(point):
    return point[2]

def make_vect(x, y):
    return ["vect", x, y]

def get_vect_x(point):
    return point[1]

def get_vect_y(point):
    return point[2]

# Napište následující funkce.
# 1. součet dvou vektorů:
# add_vects(make_vect(0, 4), make_vect(3, 0)) => ["vect", 3, 4]
# 2. násobení vektoru skalárem:
# mult_scalar_vect(2, make_vect(3, 4)) => make_vect(6, 8)
# 3. součet bodu a vektoru:
# add_point_vect(make_point(1, 1), make_vect(3, 4)) => make_point(4, 5)
# 4. velikost vektoru:
# get_vect_length(make_vect(3, 4)) => 5
# 5. rozdíl dvou bodů:
# sub_points(make_point(1, 1), make_point(4, 5)) => ["vect", 3, 4]
# 6. vzdálenost bodů:
# get_points_distance(make_point(1, 1), make_point(4, 5)) => 5


